import asyncio
import collections
import contextlib
import inspect
import logging
import typing
from abc import ABC
from functools import partial, wraps

logger = logging.getLogger(__name__)


class Function(ABC):
    def __get__(self, instance, instancetype):
        return partial(self.__call__, instance)


def run_in_threadpool_deco(func):
    @wraps(func)
    def _func(*args, **kwargs):
        return run_in_executor(None, func, *args, **kwargs)

    return _func


def run_in_threadpool(func, *args, **kwargs):
    return run_in_executor(None, func, *args, **kwargs)


def run_in_executor_deco(pool):
    def decorator(func):
        @wraps(func)
        def _func(*args, **kwargs):
            return run_in_executor(pool, func, *args, **kwargs)

        return _func

    return decorator


def run_in_executor(pool, func, *args, **kwargs):
    loop = asyncio.get_running_loop()
    if inspect.isgeneratorfunction(func):
        f = Cancellable(func(*args, **kwargs))
        future = loop.run_in_executor(pool, f)
        return CancellableFuture(future, f.cancel, loop=loop)
    else:
        return loop.run_in_executor(pool, partial(func, *args, **kwargs))


class Cancellable:
    def __init__(self, generator):
        self.generator = generator
        self.cancelled = False

    def __call__(self):
        try:
            while True:
                next(self.generator)
                if self.cancelled:
                    raise asyncio.CancelledError
        except StopIteration as e:
            return e.value

    def cancel(self):
        self.cancelled = True


class CancellableFuture(asyncio.Future):
    def __init__(self, future: asyncio.Future, cancel_func, *, loop=None):
        super().__init__(loop=loop)
        self.__future = future
        self.__cancel_func = cancel_func
        self.__future.add_done_callback(self.__future_done)

    def __future_done(self, fut: asyncio.Future):
        exc = fut.exception()
        if exc is not None:
            self.set_exception(exc)
            return

        self.set_result(self.__future.result())

    def cancel(self, msg=None):
        self.__cancel_func()
        self.__future.cancel(msg=msg)
        return super().cancel(msg=msg)


class LastManStanding:
    class __Defeat(Exception):
        pass

    def __init__(self):
        self.__locks = collections.defaultdict(asyncio.Lock)
        self.__counter = collections.defaultdict(int)

    @contextlib.asynccontextmanager
    async def join(self, key):
        with contextlib.suppress(LastManStanding.__Defeat):
            yield self.__wait(key)

    @contextlib.asynccontextmanager
    async def __wait(self, key):
        self.__counter[key] += 1
        async with self.__locks[key]:
            self.__counter[key] -= 1
            if self.__counter[key]:
                raise LastManStanding.__Defeat
            else:
                yield


class SynchronousQueue(object):
    def __init__(self):
        self._queue = asyncio.Queue(1)
        self._lock = asyncio.Lock()

    def done(self):
        self._queue.task_done()

    async def get(self):
        return await self._queue.get()

    async def put(self, item):
        async with self._lock:
            await self._queue.put(item)
            await self._queue.join()


class AsyncObjectWrapper:
    def __init__(self, obj):
        self.__obj = obj

    def __getattr__(self, name: str):
        attr = getattr(self.__obj, name)

        if name.startswith("__") and name.endswith("__"):
            return attr

        if (
            callable(attr)
            and not inspect.iscoroutinefunction(attr)
            and not inspect.isasyncgenfunction(attr)
        ):
            method = run_in_threadpool_deco(attr)
            setattr(self, name, method)
            return method

        return attr


class AsyncContextManagerWrapper:
    def __init__(
        self, context_manager: typing.ContextManager, object_wrapping: bool = False
    ):
        self._context_manager = context_manager
        self._object_wrapping = object_wrapping

    async def __aenter__(self):
        ret = await run_in_threadpool(self._context_manager.__enter__)
        return AsyncObjectWrapper(ret) if self._object_wrapping else ret

    async def __aexit__(self, *args):
        await run_in_threadpool(self._context_manager.__exit__, *args)



import threading
from abc import ABC, abstractmethod
from collections import deque


class BaseQueueIO(ABC):
    _exc: Exception

    def __init__(self):
        self._data_available = threading.Condition(threading.Lock())
        self._eof = False
        self._exc = None
        self.closed = False

    @abstractmethod
    def _write(self, data: bytes) -> None:
        """write data"""

    def write(self, data: bytes) -> None:
        with self._data_available:
            self._write(data)
            self._data_available.notify_all()

    @abstractmethod
    def _read(self, size: int = -1) -> bytes | None:
        """read data"""

    def read(self, size: int = -1) -> bytes:
        with self._data_available:
            while True:
                if self._exc is not None:
                    raise self._exc

                data = self._read(size)

                if data is None:
                    if self._eof:
                        return b""
                    self._data_available.wait()
                else:
                    return data

    def close(self):
        with self._data_available:
            self._eof = True
            self._data_available.notify_all()

    def seekable(self):
        return False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def exc(self, exc):
        with self._data_available:
            self._exc = exc
            self._data_available.notify_all()


class PopIO(BaseQueueIO):
    def __init__(self):
        super().__init__()
        self._buffer = deque()

    def _write(self, data: bytes) -> None:
        self._buffer.append(data)

    def pop(self):
        while self._buffer:
            yield self._buffer.popleft()

    def _read(self, size: int = -1) -> bytes | None:
        ret = []
        length = 0

        while self._buffer and (size < 0 or length < size):
            item = self._buffer.popleft()
            length += len(item)
            ret.append(item)

        if not ret:
            return None

        if length <= size or size < 0:
            return b"".join(ret)

        mem = memoryview(ret[-1])
        last, remain = mem[: size - length], mem[size - length :]
        self._buffer.appendleft(remain)
        ret[-1] = last
        return b"".join(ret)

    def tell(self):
        pass


class QueueIO(BaseQueueIO):
    def __init__(self):
        super().__init__()
        self._buffer = bytearray()

    def _write(self, data: bytes) -> None:
        self._buffer.extend(data)

    def _read(self, size: int = -1) -> bytes | None:
        if size < 0:
            size = len(self._buffer)

        ret = bytes(memoryview(self._buffer)[:size])

        if not ret:
            return None

        del self._buffer[:size]
        return ret



from concurrent.futures import ThreadPoolExecutor

from .asyncio import run_in_executor
from .utils import DummyStopIteration, next_without_stop_iteration


def thread_zip(*iterators):
    with ThreadPoolExecutor(len(iterators)) as p:
        iterators = [iter(iterator) for iterator in iterators]
        prefetched = [p.submit(next, iterator) for iterator in iterators]
        while True:
            try:
                rets = [task.result() for task in prefetched]
            except StopIteration:
                break
            prefetched = [p.submit(next, iterator) for iterator in iterators]
            yield rets


async def athread_zip(*iterators):
    with ThreadPoolExecutor(len(iterators)) as p:
        iterators = [iter(iterator) for iterator in iterators]
        prefetched = [
            run_in_executor(p, next_without_stop_iteration, iterator)
            for iterator in iterators
        ]
        while True:
            try:
                rets = [await task for task in prefetched]
            except DummyStopIteration:
                break
            prefetched = [
                run_in_executor(p, next_without_stop_iteration, iterator)
                for iterator in iterators
            ]
            yield rets
